#include "debug.h"

#include "FT18_STW_INIT.h"

#include <DueFlashStorage.h>

DueFlashStorage Flash;

void flash_dump_write_fault(FaultType fault_type, const uint32_t *stack,
                            const FaultStatusRegisters *fsr) {
  FlashDumpInfo info;
  memcpy(&info, flash_dump_get_info(), sizeof(info));

  FlashDump dump;
  dump.type = fault_type;
  memcpy(dump.stacked_registers, stack, sizeof(dump.stacked_registers));
  memcpy(&dump.fsr, fsr, sizeof(dump.fsr));

  uint32_t new_addr = FLASH_DUMP_ADDR_DUMPS + info.n_dumps * sizeof(dump);
  Flash.write(new_addr, (uint8_t *)&dump, sizeof(dump));

  info.n_dumps++;
  Flash.write(FLASH_DUMP_ADDR_INFO, (uint8_t *)&info, sizeof(info));
}

const FlashDumpInfo *flash_dump_get_info() {
  if (Flash.read(FLASH_DUMP_ADDR_INFO) != 0) {
    FlashDumpInfo info;
    info._zero = 0;
    info.n_dumps = 0;
    Flash.write(FLASH_DUMP_ADDR_INFO, (uint8_t *)&info, sizeof(info));
  }
  return (const FlashDumpInfo *)Flash.readAddress(FLASH_DUMP_ADDR_INFO);
}

const FlashDump *flash_dump_get_fault(uint32_t n) {
  uint32_t addr = FLASH_DUMP_ADDR_DUMPS + n * sizeof(FlashDump);
  return (const FlashDump *)Flash.readAddress(addr);
}

void uart_wait_for_txrdy() {
  while ((UART->UART_SR & UART_SR_TXRDY) != UART_SR_TXRDY)
    ;
}

size_t uart_write(uint8_t c) {
  uart_wait_for_txrdy();
  UART->UART_THR = c;
  uart_wait_for_txrdy();
  return 1;
}

size_t uart_print(const char *str) {
  const char *c = str;
  while (*c != '\0') {
    uart_write(*c);
    c++;
  }
  return c - str;
}

size_t uart_print_hex(uint32_t x) {
  for (int i = 0; i < 8; i++) {
    uint8_t nibble = (x >> (7 - i) * 4) & 0xF;
    if (nibble < 0xA) {
      uart_write(nibble + '0');
    } else {
      uart_write(nibble - 0xA + 'A');
    }
  }
  return 8;
}

size_t write(uint8_t c, bool in_irq) {
  if (in_irq) {
    uart_write(c);
  } else {
    Serial.write(c);
  }
}
size_t print(const char *str, bool in_irq) {
  if (in_irq) {
    uart_print(str);
  } else {
    Serial.print(str);
  }
}
size_t print_hex(uint32_t x, bool in_irq) {
  if (in_irq) {
    uart_print_hex(x);
  } else {
    Serial.print(x, HEX);
  }
}

void print_dumped_faults(bool in_irq) {
  uint32_t n = flash_dump_get_info()->n_dumps;

  print("====================\n", in_irq);
  print_hex(n, in_irq);
  print(" DUMPED FAULTS\n", in_irq);

  for (uint32_t i = 0; i < n; i++) {
    print("--------------------\n", in_irq);
    print("Fault ", in_irq);
    print_hex(i, in_irq);
    write('\n', in_irq);
    const FlashDump *dump = flash_dump_get_fault(i);
    print(get_fault_type_name(dump->type), in_irq);
    write('\n', in_irq);
    print_stacked_registers(dump->stacked_registers, in_irq);
    print_fault_registers(&dump->fsr, in_irq);
  }
  print("====================\n", in_irq);
}

void print_stacked_registers(const uint32_t *stack, bool in_irq) {
  print("R0    = ", in_irq);
  print_hex(stack[0], in_irq);
  write('\n', in_irq);
  print("R1    = ", in_irq);
  print_hex(stack[1], in_irq);
  write('\n', in_irq);
  print("R2    = ", in_irq);
  print_hex(stack[2], in_irq);
  write('\n', in_irq);
  print("R3    = ", in_irq);
  print_hex(stack[3], in_irq);
  write('\n', in_irq);
  print("R12   = ", in_irq);
  print_hex(stack[4], in_irq);
  write('\n', in_irq);
  print("LR    = ", in_irq);
  print_hex(stack[5], in_irq);
  write('\n', in_irq);
  print("PC    = ", in_irq);
  print_hex(stack[6], in_irq);
  write('\n', in_irq);
  print("PSR   = ", in_irq);
  print_hex(stack[7], in_irq);
  write('\n', in_irq);
  write('\n', in_irq);
}

void print_fault_registers(const FaultStatusRegisters *fsr, bool in_irq) {
  print("HFSR  = ", in_irq);
  print_hex(fsr->HFSR, in_irq);
  write('\n', in_irq);
  print("CFSR  = ", in_irq);
  print_hex(fsr->CFSR, in_irq);
  write('\n', in_irq);
  print("MMFAR = ", in_irq);
  print_hex(fsr->MMFAR, in_irq);
  write('\n', in_irq);
  print("BFAR  = ", in_irq);
  print_hex(fsr->BFAR, in_irq);
  write('\n', in_irq);

  print("DFSR  = ", in_irq);
  print_hex(fsr->DFSR, in_irq);
  write('\n', in_irq);
  print("AFSR  = ", in_irq);
  print_hex(fsr->AFSR, in_irq);
  write('\n', in_irq);
  print("SHCSR = ", in_irq);
  print_hex(fsr->SHCSR, in_irq);
  write('\n', in_irq);
  write('\n', in_irq);
}

FaultStatusRegisters get_current_fsr() {
  FaultStatusRegisters fsr;
  fsr.HFSR = SCB->HFSR;
  fsr.CFSR = SCB->CFSR;
  fsr.MMFAR = SCB->MMFAR;
  fsr.BFAR = SCB->BFAR;

  fsr.DFSR = SCB->DFSR;
  fsr.AFSR = SCB->AFSR;
  fsr.SHCSR = SCB->SHCSR;
  return fsr;
}

const char *get_fault_type_name(FaultType type) {
  switch (type) {
  case FaultType::HardFault:
    return "HARD FAULT";
  case FaultType::MemManage:
    return "MEMMANGAGE FAULT";
  case FaultType::BusFault:
    return "BUS FAULT";
  case FaultType::UsageFault:
    return "USAGE FAULT";
  }
}

void fault_handler(uint32_t *stack_addr, FaultType fault_type, const int *leds,
                   unsigned n_leds) {
  Serial.begin(9600);
  for (unsigned i = 0; i < n_leds; i++) {
    pinMode(leds[i], OUTPUT);
  }

  FaultStatusRegisters fsr = get_current_fsr();

  flash_dump_write_fault(fault_type, stack_addr, &fsr);

  while (1) {
    uart_print("====================\n");
    uart_print(get_fault_type_name(fault_type));
    uart_print("\n====================\n");

    print_stacked_registers(stack_addr, true);
    print_fault_registers(&fsr, true);
    for (unsigned i = 0; i < n_leds; i++) {
      digitalWrite(leds[i], HIGH);
    }
    busy_wait(5000000);
    for (unsigned i = 0; i < n_leds; i++) {
      digitalWrite(leds[i], LOW);
    }
    busy_wait(5000000);
  }
}

// If we don't wrap the handler functions in an extern C block, their names get
// mangled and they won't override the weakly-linked handler functions defined
// in the Arduino(-SAM) framework.
extern "C" {

// Get the stack pointer and pass it to the respective C handler. Depending on
// the EXC_RETURN value, we need to either use the Main Stack Pointer (msp) or
// the Process Stack Pointer (psp).
void HardFault_Handler(void) {
  __ASM("tst lr, #4\n"
        "ite eq\n"
        "mrseq r0, msp\n"
        "mrsne r0, psp\n"
        "b HardFault_Handler_C");
}

void MemManage_Handler(void) {
  __ASM("tst lr, #4\n"
        "ite eq\n"
        "mrseq r0, msp\n"
        "mrsne r0, psp\n"
        "b MemManage_Handler_C");
}

void BusFault_Handler(void) {
  __ASM("tst lr, #4\n"
        "ite eq\n"
        "mrseq r0, msp\n"
        "mrsne r0, psp\n"
        "b BusFault_Handler_C");
}

void UsageFault_Handler(void) {
  __ASM("tst lr, #4\n"
        "ite eq\n"
        "mrseq r0, msp\n"
        "mrsne r0, psp\n"
        "b UsageFault_Handler_C");
}

void HardFault_Handler_C(uint32_t *stack_addr) {
  // All LEDs except for the bottom ones
  static const int leds[] = {led1, led2, led3,  led4,  led5,  led6,  led7,
                             led8, led9, led10, led11, led12, led14, led15};
  fault_handler(stack_addr, FaultType::HardFault, leds,
                sizeof(leds) / sizeof(int));
}

void MemManage_Handler_C(uint32_t *stack_addr) {
  // Left of display
  static const int leds[] = {led11, led12, led13};
  fault_handler(stack_addr, FaultType::MemManage, leds,
                sizeof(leds) / sizeof(int));
}

void BusFault_Handler_C(uint32_t *stack_addr) {
  // Right of display
  static const int leds[] = {led14, led15, led16};
  fault_handler(stack_addr, FaultType::BusFault, leds,
                sizeof(leds) / sizeof(int));
}

void UsageFault_Handler_C(uint32_t *stack_addr) {
  // Left and right of display
  static const int leds[] = {led11, led12, led13, led14, led15, led16};
  fault_handler(stack_addr, FaultType::UsageFault, leds,
                sizeof(leds) / sizeof(int));
}
}
